const TYPE_ENUM = {
  i8: 'i8',
  ui8: 'ui8',
  ui8c: 'ui8c',
  i16: 'i16',
  ui16: 'ui16',
  i32: 'i32',
  ui32: 'ui32',
  f32: 'f32',
  f64: 'f64'
};
const TYPE_NAME = {
  i8: 'Int8',
  ui8: 'Uint8',
  ui8c: 'Uint8Clamped',
  i16: 'Int16',
  ui16: 'Uint16',
  i32: 'Int32',
  ui32: 'Uint32',
  f32: 'Float32',
  f64: 'Float64'
};
const TYPES = {
  i8: Int8Array,
  ui8: Uint8Array,
  ui8c: Uint8ClampedArray,
  i16: Int16Array,
  ui16: Uint16Array,
  i32: Int32Array,
  ui32: Uint32Array,
  f32: Float32Array,
  f64: Float64Array
};
const UNSIGNED_MAX = {
  uint8: 2 ** 8,
  uint16: 2 ** 16,
  uint32: 2 ** 32
};

const roundToMultiple4 = x => Math.ceil(x / 4) * 4;

const $storeRef = Symbol('storeRef');
const $storeSize = Symbol('storeSize');
const $storeMaps = Symbol('storeMaps');
const $storeFlattened = Symbol('storeFlattened');
const $storeBase = Symbol('storeBase');
const $storeType = Symbol('storeType');
const $storeArrayCounts = Symbol('storeArrayCount');
const $storeSubarrays = Symbol('storeSubarrays');
const $storeCursor = Symbol('storeCursor');
const $subarrayCursors = Symbol('subarrayCursors');
const $subarray = Symbol('subarray');
const $tagStore = Symbol('tagStore');
const $queryShadow = Symbol('queryShadow');
const $serializeShadow = Symbol('serializeShadow');
const $indexType = Symbol('indexType');
const $indexBytes = Symbol('indexBytes');
const stores = {};
const resize = (ta, size) => {
  const newBuffer = new SharedArrayBuffer(size * ta.BYTES_PER_ELEMENT);
  const newTa = new ta.constructor(newBuffer);
  newTa.set(ta, 0);
  return newTa;
};

const resizeSubarray = (metadata, store, size) => {
  const cursors = metadata[$subarrayCursors];
  const type = store[$storeType];
  const length = store[0].length;
  const indexType = length < UNSIGNED_MAX.uint8 ? 'ui8' : length < UNSIGNED_MAX.uint16 ? 'ui16' : 'ui32';
  const arrayCount = metadata[$storeArrayCounts][type];
  const summedLength = Array(arrayCount).fill(0).reduce((a, p) => a + length, 0); // for threaded impl

  const summedBytesPerElement = Array(arrayCount).fill(0).reduce((a, p) => a + TYPES[type].BYTES_PER_ELEMENT, 0);
  const totalBytes = roundToMultiple4(summedBytesPerElement * summedLength * size);
  const buffer = new SharedArrayBuffer(totalBytes);
  const array = new TYPES[type](buffer);
  array.set(metadata[$storeSubarrays][type]);
  metadata[$storeSubarrays][type] = array;
  metadata[$storeSubarrays][type][$queryShadow] = array.slice(0);
  metadata[$storeSubarrays][type][$serializeShadow] = array.slice(0);
  array[$indexType] = TYPE_NAME[indexType];
  array[$indexBytes] = TYPES[indexType].BYTES_PER_ELEMENT;
  let end = 0;

  for (let eid = 0; eid < size; eid++) {
    const from = cursors[type] + eid * length;
    const to = from + length;
    store[eid] = metadata[$storeSubarrays][type].subarray(from, to);
    store[eid].from = from;
    store[eid].to = to;
    store[eid][$queryShadow] = metadata[$storeSubarrays][type][$queryShadow].subarray(from, to);
    store[eid][$serializeShadow] = metadata[$storeSubarrays][type][$serializeShadow].subarray(from, to);
    store[eid][$subarray] = true;
    store[eid][$indexType] = array[$indexType];
    store[eid][$indexBytes] = array[$indexBytes];
    end = to;
  }

  cursors[type] = end;
};

const resizeRecursive = (metadata, store, size) => {
  Object.keys(store).forEach(key => {
    const ta = store[key];

    if (Array.isArray(ta)) {
      resizeSubarray(metadata, ta, size);
      store[$storeFlattened].push(ta);
    } else if (ArrayBuffer.isView(ta)) {
      store[key] = resize(ta, size);
      store[$storeFlattened].push(store[key]);
      store[key][$queryShadow] = resize(ta[$queryShadow], size);
      store[key][$serializeShadow] = resize(ta[$serializeShadow], size);
    } else if (typeof ta === 'object') {
      resizeRecursive(metadata, store[key], size);
    }
  });
};

const resizeStore = (store, size) => {
  if (store[$tagStore]) return;
  store[$storeSize] = size;
  store[$storeFlattened].length = 0;
  Object.keys(store[$subarrayCursors]).forEach(k => {
    store[$subarrayCursors][k] = 0;
  });
  resizeRecursive(store, store, size); // resizeSubarrays(store, size)
};
const resetStoreFor = (store, eid) => {
  if (store[$storeFlattened]) {
    store[$storeFlattened].forEach(ta => {
      if (ArrayBuffer.isView(ta)) ta[eid] = 0;else ta[eid].fill(0);
    });
  }
};

const createTypeStore = (type, length) => {
  const totalBytes = length * TYPES[type].BYTES_PER_ELEMENT;
  const buffer = new SharedArrayBuffer(totalBytes);
  return new TYPES[type](buffer);
};

const createArrayStore = (metadata, type, length) => {
  const size = metadata[$storeSize];
  const store = Array(size).fill(0);
  store[$storeType] = type;
  const cursors = metadata[$subarrayCursors];
  const indexType = length < UNSIGNED_MAX.uint8 ? 'ui8' : length < UNSIGNED_MAX.uint16 ? 'ui16' : 'ui32';
  if (!length) throw new Error('❌ Must define a length for component array.');
  if (!TYPES[type]) throw new Error(`❌ Invalid component array property type ${type}.`); // create buffer for type if it does not already exist

  if (!metadata[$storeSubarrays][type]) {
    const arrayCount = metadata[$storeArrayCounts][type];
    const summedLength = Array(arrayCount).fill(0).reduce((a, p) => a + length, 0); // for threaded impl

    const summedBytesPerElement = Array(arrayCount).fill(0).reduce((a, p) => a + TYPES[type].BYTES_PER_ELEMENT, 0);
    const totalBytes = roundToMultiple4(summedBytesPerElement * summedLength * size);
    const buffer = new SharedArrayBuffer(totalBytes);
    const array = new TYPES[type](buffer);
    metadata[$storeSubarrays][type] = array;
    metadata[$storeSubarrays][type][$queryShadow] = array.slice(0);
    metadata[$storeSubarrays][type][$serializeShadow] = array.slice(0);
    array[$indexType] = TYPE_NAME[indexType];
    array[$indexBytes] = TYPES[indexType].BYTES_PER_ELEMENT;
  } // pre-generate subarrays for each eid


  let end = 0;

  for (let eid = 0; eid < size; eid++) {
    const from = cursors[type] + eid * length;
    const to = from + length;
    store[eid] = metadata[$storeSubarrays][type].subarray(from, to);
    store[eid].from = from;
    store[eid].to = to;
    store[eid][$queryShadow] = metadata[$storeSubarrays][type][$queryShadow].subarray(from, to);
    store[eid][$serializeShadow] = metadata[$storeSubarrays][type][$serializeShadow].subarray(from, to);
    store[eid][$subarray] = true;
    store[eid][$indexType] = TYPE_NAME[indexType];
    store[eid][$indexBytes] = TYPES[indexType].BYTES_PER_ELEMENT;
    end = to;
  }

  cursors[type] = end;
  return store;
};

const createShadows = store => {
  store[$queryShadow] = store.slice(0);
  store[$serializeShadow] = store.slice(0);
};

const isArrayType = x => Array.isArray(x) && typeof x[0] === 'string' && typeof x[1] === 'number';

const createStore = (schema, size) => {
  const $store = Symbol('store');

  if (!schema || !Object.keys(schema).length) {
    // tag component
    stores[$store] = {
      [$storeSize]: size,
      [$tagStore]: true,
      [$storeBase]: () => stores[$store]
    };
    return stores[$store];
  }

  schema = JSON.parse(JSON.stringify(schema));
  const arrayCounts = {};

  const collectArrayCounts = s => {
    const keys = Object.keys(s);

    for (const k of keys) {
      if (isArrayType(s[k])) {
        if (!arrayCounts[s[k][0]]) arrayCounts[s[k][0]] = 0;
        arrayCounts[s[k][0]]++;
      } else if (s[k] instanceof Object) {
        collectArrayCounts(s[k]);
      }
    }
  };

  collectArrayCounts(schema);
  const metadata = {
    [$storeSize]: size,
    [$storeMaps]: {},
    [$storeSubarrays]: {},
    [$storeRef]: $store,
    [$storeCursor]: 0,
    [$subarrayCursors]: Object.keys(TYPES).reduce((a, type) => ({ ...a,
      [type]: 0
    }), {}),
    [$storeFlattened]: [],
    [$storeArrayCounts]: arrayCounts
  };

  if (schema instanceof Object && Object.keys(schema).length) {
    const recursiveTransform = (a, k) => {
      if (typeof a[k] === 'string') {
        a[k] = createTypeStore(a[k], size);
        createShadows(a[k]);

        a[k][$storeBase] = () => stores[$store];

        metadata[$storeFlattened].push(a[k]);
      } else if (isArrayType(a[k])) {
        const [type, length] = a[k];
        a[k] = createArrayStore(metadata, type, length);

        a[k][$storeBase] = () => stores[$store];

        metadata[$storeFlattened].push(a[k]);
        Object.seal(a[k]);
      } else if (a[k] instanceof Object) {
        a[k] = Object.keys(a[k]).reduce(recursiveTransform, a[k]);
        Object.seal(a[k]);
      }

      return a;
    };

    stores[$store] = Object.assign(Object.keys(schema).reduce(recursiveTransform, schema), metadata);

    stores[$store][$storeBase] = () => stores[$store];

    Object.seal(stores[$store]);
    return stores[$store];
  }
};

let resized = false;
const setSerializationResized = v => {
  resized = v;
};

const canonicalize = target => {
  let componentProps = [];
  let changedProps = new Set();

  if (Array.isArray(target)) {
    componentProps = target.map(p => {
      if (typeof p === 'function' && p.name === 'QueryChanged') {
        p()[$storeFlattened].forEach(prop => {
          changedProps.add(prop);
        });
        return p()[$storeFlattened];
      }

      if (Object.getOwnPropertySymbols(p).includes($storeFlattened)) {
        return p[$storeFlattened];
      }

      if (Object.getOwnPropertySymbols(p).includes($storeBase)) {
        return p;
      }
    }).reduce((a, v) => a.concat(v), []);
  }

  return [componentProps, changedProps];
};

const defineSerializer = (target, maxBytes = 20000000) => {
  const isWorld = Object.getOwnPropertySymbols(target).includes($componentMap);
  let [componentProps, changedProps] = canonicalize(target); // TODO: calculate max bytes based on target

  const buffer = new ArrayBuffer(maxBytes);
  const view = new DataView(buffer); // const write = defineWriter(view, componentProps, changedProps)

  return ents => {
    if (resized) {
      [componentProps, changedProps] = canonicalize(target);
      resized = false;
    }

    if (isWorld) {
      componentProps = [];
      target[$componentMap].forEach((c, component) => {
        if (component[$storeFlattened]) componentProps.push(...component[$storeFlattened]);else componentProps.push(component);
      });
    }

    let world;

    if (Object.getOwnPropertySymbols(ents).includes($componentMap)) {
      world = ents;
      ents = ents[$entityArray];
    } else {
      world = eidToWorld.get(ents[0]);
    }

    if (!ents.length) return; // view.where = 0
    // // save space for entity count
    // const countWhere = where
    // view.where += 4
    // // write eid,val,val,val..., eid,val,val...
    // for (let i = 0; i < ents.length; i++) {
    //   const eid = ents[i]
    //   write(world, eid)
    // }
    // // view.setUint32(countWhere, count)
    // const end = view.where
    // view.where = 0
    // return buffer.slice(0, end)

    let where = 0; // iterate over component props

    for (let pid = 0; pid < componentProps.length; pid++) {
      const prop = componentProps[pid];
      const diff = changedProps.has(prop); // write pid

      view.setUint8(where, pid);
      where += 1; // save space for entity count

      const countWhere = where;
      where += 4;
      let count = 0; // write eid,val

      for (let i = 0; i < ents.length; i++) {
        const eid = ents[i]; // skip if entity doesn't have this component

        if (!hasComponent(world, prop[$storeBase](), eid)) {
          continue;
        } // skip if diffing and no change


        if (diff && prop[eid] === prop[$serializeShadow][eid]) {
          continue;
        }

        count++; // write eid

        view.setUint32(where, eid);
        where += 4;

        if (prop[$tagStore]) {
          continue;
        } // if property is an array


        if (ArrayBuffer.isView(prop[eid])) {
          const type = prop[eid].constructor.name.replace('Array', '');
          const indexType = prop[eid][$indexType];
          const indexBytes = prop[eid][$indexBytes]; // add space for count of dirty array elements

          const countWhere2 = where;
          where += 1;
          let count2 = 0; // write index,value

          for (let i = 0; i < prop[eid].length; i++) {
            const value = prop[eid][i];

            if (diff && prop[eid][i] === prop[eid][$serializeShadow][i]) {
              continue;
            } // write array index


            view[`set${indexType}`](where, i);
            where += indexBytes; // write value at that index

            view[`set${type}`](where, value);
            where += prop[eid].BYTES_PER_ELEMENT;
            count2++;
          } // write total element count


          view[`set${indexType}`](countWhere2, count2);
        } else {
          // regular property values
          const type = prop.constructor.name.replace('Array', ''); // set value next [type] bytes

          view[`set${type}`](where, prop[eid]);
          where += prop.BYTES_PER_ELEMENT; // sync shadow state

          prop[$serializeShadow][eid] = prop[eid];
        }
      }

      view.setUint32(countWhere, count);
    }

    return buffer.slice(0, where);
  };
};
const defineDeserializer = target => {
  const isWorld = Object.getOwnPropertySymbols(target).includes($componentMap);
  let [componentProps] = canonicalize(target);
  return (world, packet) => {
    if (resized) {
      [componentProps] = canonicalize(target);
      resized = false;
    }

    if (isWorld) {
      componentProps = [];
      target[$componentMap].forEach((c, component) => {
        if (component[$storeFlattened]) componentProps.push(...component[$storeFlattened]);else componentProps.push(component);
      });
    }

    const view = new DataView(packet);
    let where = 0;
    const newEntities = new Map();

    while (where < packet.byteLength) {
      // pid
      const pid = view.getUint8(where);
      where += 1; // entity count

      const entityCount = view.getUint32(where);
      where += 4; // component property

      const prop = componentProps[pid]; // Get the entities and set their prop values

      for (let i = 0; i < entityCount; i++) {
        let eid = view.getUint32(where);
        where += 4;
        let newEid = newEntities.get(eid);

        if (newEid !== undefined) {
          eid = newEid;
        } // if this world hasn't seen this eid yet


        if (!world[$entityEnabled][eid]) {
          // make a new entity for the data
          const newEid = addEntity(world);
          newEntities.set(eid, newEid);
          eid = newEid;
        }

        const component = prop[$storeBase]();

        if (!hasComponent(world, component, eid)) {
          addComponent(world, component, eid);
        }

        if (component[$tagStore]) {
          continue;
        }

        if (ArrayBuffer.isView(prop[eid])) {
          const array = prop[eid];
          const count = view[`get${array[$indexType]}`](where);
          where += array[$indexBytes]; // iterate over count

          for (let i = 0; i < count; i++) {
            const index = view[`get${array[$indexType]}`](where);
            where += array[$indexBytes];
            const value = view[`get${array.constructor.name.replace('Array', '')}`](where);
            where += array.BYTES_PER_ELEMENT;
            prop[eid][index] = value;
          }
        } else {
          const value = view[`get${prop.constructor.name.replace('Array', '')}`](where);
          where += prop.BYTES_PER_ELEMENT;
          prop[eid] = value;
        }
      }
    }
  };
};

const $entityMasks = Symbol('entityMasks');
const $entityEnabled = Symbol('entityEnabled');
const $entityArray = Symbol('entityArray');
const $entityIndices = Symbol('entityIndices');
const NONE$1 = 2 ** 32;
const defaultSize = 100000; // need a global EID cursor which all worlds and all components know about
// so that world entities can posess entire rows spanning all component tables

let globalEntityCursor = 0;
let globalSize = defaultSize;

let resizeThreshold = () => globalSize - globalSize / 5;

const getGlobalSize = () => globalSize; // removed eids should also be global to prevent memory leaks

const removed = [];
const getEntityCursor = () => globalEntityCursor;
const eidToWorld = new Map();
const addEntity = world => {
  const enabled = world[$entityEnabled];
  const eid = removed.length > 0 ? removed.shift() : globalEntityCursor++;
  enabled[eid] = 1;
  world[$entityIndices][eid] = world[$entityArray].push(eid) - 1;
  eidToWorld.set(eid, world); // if data stores are 80% full

  if (globalEntityCursor >= resizeThreshold()) {
    // grow by half the original size rounded up to a multiple of 4
    const size = globalSize;
    const amount = Math.ceil(size / 2 / 4) * 4;
    const newSize = size + amount;
    globalSize = newSize;
    resizeWorlds(newSize);
    resizeComponents(newSize);
    setSerializationResized(true);
    console.info(`👾 bitECS - resizing all worlds from ${size} to ${size + amount}`);
  }

  return eid;
};

const popSwap = (world, eid) => {
  // pop swap
  const index = world[$entityIndices][eid];
  const swapped = world[$entityArray].pop();

  if (swapped !== eid) {
    world[$entityArray][index] = swapped;
    world[$entityIndices][swapped] = index;
  }

  world[$entityIndices][eid] = NONE$1;
};

const removeEntity = (world, eid) => {
  const enabled = world[$entityEnabled]; // Check if entity is already removed

  if (enabled[eid] === 0) return; // Remove entity from all queries
  // TODO: archetype graph

  world[$queries].forEach(query => {
    queryRemoveEntity(world, query, eid);
  }); // Free the entity

  removed.push(eid);
  enabled[eid] = 0; // pop swap

  popSwap(world, eid); // Clear entity bitmasks

  for (let i = 0; i < world[$entityMasks].length; i++) world[$entityMasks][i][eid] = 0;
};

function Not(c) {
  return function QueryNot() {
    return c;
  };
}
function Changed(c) {
  return function QueryChanged() {
    return c;
  };
}
const $queries = Symbol('queries');
const $queryMap = Symbol('queryMap');
const $dirtyQueries = Symbol('$dirtyQueries');
const $queryComponents = Symbol('queryComponents');
const NONE = 2 ** 32;
const enterQuery = query => world => {
  if (!world[$queryMap].has(query)) registerQuery(world, query);
  const q = world[$queryMap].get(query);
  return q.entered.splice(0);
};
const exitQuery = query => world => {
  if (!world[$queryMap].has(query)) registerQuery(world, query);
  const q = world[$queryMap].get(query);
  return q.exited.splice(0);
};
const registerQuery = (world, query) => {
  let components = [];
  let notComponents = [];
  let changedComponents = [];
  query[$queryComponents].forEach(c => {
    if (typeof c === 'function') {
      if (c.name === 'QueryNot') {
        notComponents.push(c());
      }

      if (c.name === 'QueryChanged') {
        changedComponents.push(c());
        components.push(c());
      }
    } else {
      components.push(c);
    }
  });

  const mapComponents = c => world[$componentMap].get(c);

  const size = components.concat(notComponents).reduce((a, c) => c[$storeSize] > a ? c[$storeSize] : a, 0);
  const entities = [];
  const changed = [];
  const indices = new Uint32Array(size).fill(NONE);
  const enabled = new Uint8Array(size);
  const generations = components.concat(notComponents).map(c => {
    if (!world[$componentMap].has(c)) registerComponent(world, c);
    return c;
  }).map(mapComponents).map(c => c.generationId).reduce((a, v) => {
    if (a.includes(v)) return a;
    a.push(v);
    return a;
  }, []);

  const reduceBitmasks = (a, c) => {
    if (!a[c.generationId]) a[c.generationId] = 0;
    a[c.generationId] |= c.bitflag;
    return a;
  };

  const masks = components.map(mapComponents).reduce(reduceBitmasks, {});
  const notMasks = notComponents.map(mapComponents).reduce((a, c) => {
    if (!a[c.generationId]) {
      a[c.generationId] = 0;
      a[c.generationId] |= c.bitflag;
    }

    return a;
  }, {});
  const flatProps = components.map(c => Object.getOwnPropertySymbols(c).includes($storeFlattened) ? c[$storeFlattened] : [c]).reduce((a, v) => a.concat(v), []);
  const toRemove = [];
  const entered = [];
  const exited = [];
  world[$queryMap].set(query, {
    entities,
    changed,
    enabled,
    components,
    notComponents,
    changedComponents,
    masks,
    notMasks,
    generations,
    indices,
    flatProps,
    toRemove,
    entered,
    exited
  });
  world[$queries].add(query);

  for (let eid = 0; eid < getEntityCursor(); eid++) {
    if (!world[$entityEnabled][eid]) continue;

    if (queryCheckEntity(world, query, eid)) {
      queryAddEntity(world, query, eid);
    }
  }
};

const diff = q => {
  q.changed.length = 0;
  const flat = q.flatProps;

  for (let i = 0; i < q.entities.length; i++) {
    const eid = q.entities[i];
    let dirty = false;

    for (let pid = 0; pid < flat.length; pid++) {
      const prop = flat[pid];

      if (ArrayBuffer.isView(prop[eid])) {
        for (let i = 0; i < prop[eid].length; i++) {
          if (prop[eid][i] !== prop[eid][$queryShadow][i]) {
            dirty = true;
            prop[eid][$queryShadow][i] = prop[eid][i];
          }
        }
      } else {
        if (prop[eid] !== prop[$queryShadow][eid]) {
          dirty = true;
          prop[$queryShadow][eid] = prop[eid];
        }
      }
    }

    if (dirty) q.changed.push(eid);
  }

  return q.changed;
};

const defineQuery = components => {
  const query = function (world) {
    if (!world[$queryMap].has(query)) registerQuery(world, query);
    const q = world[$queryMap].get(query); // queryHooks(q)

    queryCommitRemovals(world, q);
    if (q.changedComponents.length) return diff(q);
    return q.entities;
  };

  query[$queryComponents] = components;
  return query;
}; // TODO: archetype graph

const queryCheckEntity = (world, query, eid) => {
  const {
    masks,
    notMasks,
    generations
  } = world[$queryMap].get(query);

  for (let i = 0; i < generations.length; i++) {
    const generationId = generations[i];
    const qMask = masks[generationId];
    const qNotMask = notMasks[generationId];
    const eMask = world[$entityMasks][generationId][eid];

    if (qNotMask && (eMask & qNotMask) !== 0) {
      return false;
    }

    if (qMask && (eMask & qMask) !== qMask) {
      return false;
    }
  }

  return true;
};
const queryCheckComponent = (world, query, component) => {
  const {
    generationId,
    bitflag
  } = world[$componentMap].get(component);
  const {
    masks
  } = world[$queryMap].get(query);
  const mask = masks[generationId];
  return (mask & bitflag) === bitflag;
};
const queryAddEntity = (world, query, eid) => {
  const q = world[$queryMap].get(query);
  if (q.enabled[eid]) return;
  q.enabled[eid] = true;
  q.entities.push(eid);
  q.indices[eid] = q.entities.length - 1; // TODO: pop swap so dupes don't enter

  q.entered.push(eid);
};

const queryCommitRemovals = (world, q) => {
  while (q.toRemove.length) {
    const eid = q.toRemove.pop();
    const index = q.indices[eid];
    if (index === NONE) continue;
    const swapped = q.entities.pop();

    if (swapped !== eid) {
      q.entities[index] = swapped;
      q.indices[swapped] = index;
    }

    q.indices[eid] = NONE;
  }

  world[$dirtyQueries].delete(q);
};

const commitRemovals = world => {
  world[$dirtyQueries].forEach(q => {
    queryCommitRemovals(world, q);
  });
};
const queryRemoveEntity = (world, query, eid) => {
  const q = world[$queryMap].get(query);
  if (!q.enabled[eid]) return;
  q.enabled[eid] = false;
  q.toRemove.push(eid);
  world[$dirtyQueries].add(q); // TODO: pop swap so dupes don't enter

  q.exited.push(eid);
};

const $componentMap = Symbol('componentMap');
const storeComponents = [];
const resizeComponents = size => {
  storeComponents.forEach(component => resizeStore(component, size));
};
const defineComponent = schema => {
  const component = createStore(schema, defaultSize);
  if (schema && Object.keys(schema).length) storeComponents.push(component);
  return component;
};
const incrementBitflag = world => {
  world[$bitflag] *= 2;

  if (world[$bitflag] >= 2 ** 32) {
    world[$bitflag] = 1;
    world[$entityMasks].push(new Uint32Array(world[$size]));
  }
};
const registerComponent = (world, component) => {
  world[$componentMap].set(component, {
    generationId: world[$entityMasks].length - 1,
    bitflag: world[$bitflag],
    store: component
  });

  if (component[$storeSize] < world[$size]) {
    resizeStore(component, world[$size]);
  }

  incrementBitflag(world);
};
const registerComponents = (world, components) => {
  components.forEach(c => registerComponent(world, c));
};
const hasComponent = (world, component, eid) => {
  const registeredComponent = world[$componentMap].get(component);
  if (!registeredComponent) return;
  const {
    generationId,
    bitflag
  } = registeredComponent;
  const mask = world[$entityMasks][generationId][eid];
  return (mask & bitflag) === bitflag;
};
const addComponent = (world, component, eid, reset = false) => {
  if (!world[$componentMap].has(component)) registerComponent(world, component);
  if (hasComponent(world, component, eid)) return; // Add bitflag to entity bitmask

  const {
    generationId,
    bitflag
  } = world[$componentMap].get(component);
  world[$entityMasks][generationId][eid] |= bitflag; // todo: archetype graph

  world[$queries].forEach(query => {
    if (!queryCheckComponent(world, query, component)) return;
    const match = queryCheckEntity(world, query, eid);
    if (match) queryAddEntity(world, query, eid);
  }); // Zero out each property value

  if (reset) resetStoreFor(component, eid);
};
const removeComponent = (world, component, eid, reset = true) => {
  const {
    generationId,
    bitflag
  } = world[$componentMap].get(component);
  if (!(world[$entityMasks][generationId][eid] & bitflag)) return; // todo: archetype graph

  world[$queries].forEach(query => {
    if (!queryCheckComponent(world, query, component)) return;
    const match = queryCheckEntity(world, query, eid);
    if (match) queryRemoveEntity(world, query, eid);
  }); // Remove flag from entity bitmask

  world[$entityMasks][generationId][eid] &= ~bitflag; // Zero out each property value

  if (reset) resetStoreFor(component, eid);
};

const $size = Symbol('size');
const $resizeThreshold = Symbol('resizeThreshold');
const $bitflag = Symbol('bitflag');
const worlds = [];
const resizeWorlds = size => {
  worlds.forEach(world => {
    world[$size] = size;
    world[$queryMap].forEach(q => {
      q.indices = resize(q.indices, size);
      q.enabled = resize(q.enabled, size);
    });
    world[$entityEnabled] = resize(world[$entityEnabled], size);
    world[$entityIndices] = resize(world[$entityIndices], size);

    for (let i = 0; i < world[$entityMasks].length; i++) {
      const masks = world[$entityMasks][i];
      world[$entityMasks][i] = resize(masks, size);
    }

    world[$resizeThreshold] = world[$size] - world[$size] / 5;
  });
};
const createWorld = () => {
  const world = {};
  const size = getGlobalSize();
  world[$size] = size;
  world[$entityEnabled] = new Uint8Array(size);
  world[$entityMasks] = [new Uint32Array(size)];
  world[$entityArray] = [];
  world[$entityIndices] = new Uint32Array(size);
  world[$bitflag] = 1;
  world[$componentMap] = new Map();
  world[$queryMap] = new Map();
  world[$queries] = new Set();
  world[$dirtyQueries] = new Set();
  worlds.push(world);
  return world;
};

const defineSystem = (fn1, fn2) => {
  const update = fn2 !== undefined ? fn2 : fn1;
  const create = fn2 !== undefined ? fn1 : undefined;
  const init = new Set();

  const system = world => {
    if (create && !init.has(world)) {
      create(world);
      init.add(world);
    }

    update(world);
    commitRemovals(world);
    return world;
  };

  Object.defineProperty(system, 'name', {
    value: (update.name || "AnonymousSystem") + "_internal",
    configurable: true
  });
  return system;
};

const pipe = (...fns) => input => {
  if (!input || Array.isArray(input) && input.length === 0) return;
  fns = Array.isArray(fns[0]) ? fns[0] : fns;
  let tmp = input;

  for (let i = 0; i < fns.length; i++) {
    const fn = fns[i];

    if (Array.isArray(tmp)) {
      tmp = tmp.reduce((a, v) => a.concat(fn(v)), []);
    } else {
      tmp = fn(tmp);
    }
  }

  return tmp;
};
const Types = TYPE_ENUM;

export { Changed, Not, Types, addComponent, addEntity, commitRemovals, createWorld, defineComponent, defineDeserializer, defineQuery, defineSerializer, defineSystem, enterQuery, exitQuery, hasComponent, pipe, registerComponent, registerComponents, removeComponent, removeEntity };
//# sourceMappingURL=index.es.js.map
